module AbsSyntax where
import Utils

data Expr a
=  EVar Name				-- Variables
| ENum Int				-- Numbers
| EConstr Int Int			-- Constructor tag arity
| EAp (Expr a) (Expr a)			-- Applications
| ELet IsRec [(a, Expr a)] (Expr a)	-- Body of let(rec)
| ECase (Expr a) [Alter a]		-- Alternatives
| ELam [a] (Expr a)             	-- Lambda abstractions
| EUnion [Expr a]			-- Abstract value union
instance Text (Expr a)
type CoreExpr = Expr Name
type Name = String
type IsRec = Bool
recursive, nonRecursive :: IsRec
recursive    = True
nonRecursive = False
bindersOf :: [(a,b)] -> [a]
bindersOf defns =  [name | (name, rhs) <- defns]
rhssOf        :: [(a,b)] -> [b]
rhssOf defns  =  [rhs  | (name, rhs) <- defns]
type Alter a = (Int, [a], Expr a)
type CoreAlt = Alter Name
isAtomicExpr :: Expr a -> Bool
isAtomicExpr (EVar v) = True
isAtomicExpr (ENum n) = True
isAtomicExpr e        = False
type Program a = [ScDefn a]
type CoreProgram = Program Name
type ScDefn a = (Name, [a], Expr a)
type CoreScDefn = ScDefn Name
preludeDefs :: CoreProgram
preludeDefs
  = [ ("I", ["x"], EVar "x"),
      ("K", ["x","y"], EVar "x"),
      ("K1",["x","y"], EVar "y"),
      ("S", ["f","g","x"], EAp (EAp (EVar "f") (EVar "x"))
                               (EAp (EVar "g") (EVar "x"))),
      ("compose", ["f","g","x"], EAp (EVar "f")
                                      (EAp (EVar "g") (EVar "x"))),
      ("twice", ["f", "x"], EAp (EAp (EAp (EVar "compose") (EVar "f")) (EVar "f")) (EVar "x")),
("if", ["c", "t", "f"], ECase (EVar "c") 
		[(1, [], EVar "f"), (2, [], EVar "t")])]
mkMultiAp :: Int -> CoreExpr -> CoreExpr -> CoreExpr
mkMultiAp n e1 e2 = foldl EAp e1 (take n e2s)
                    where
                    e2s = e2 : e2s
iNil     :: Iseq                  -- The empty iseq
iStr     :: String -> Iseq        -- Turn a string into an iseq
iAppend  :: Iseq -> Iseq -> Iseq  -- Append two iseqs
iNewline :: Iseq                  -- New line with indentation
iIndent  :: Iseq -> Iseq          -- Indent an iseq
iDisplay :: Iseq -> String        -- Turn an iseq into a string
pprExpr :: CoreExpr -> Iseq
pprExpr (ENum n) = iStr (show n)
pprExpr (EVar v) = iStr v
pprExpr (EConstr t a) = iConcat [iStr "Cons{", iNum t, iStr ",", 
                                  iNum a, iStr "}"]
pprExpr (EAp e1 e2) = (pprExpr e1) `iAppend` 
						((iStr " ") `iAppend` (pprParExpr e2))
    where
  pprParExpr e@(EAp _ _) = iConcat [iStr "(", pprExpr e, iStr ")"]
  pprParExpr e           = pprExpr e
pprExpr (ECase e alts)
  = iConcat [  iStr "case ", pprExpr e, iStr " of", iNewline,
               iStr "  ", iIndent (iInterleave iNl (map pprAlt alts)) ]
    where
    iNl = iConcat [ iStr ";", iNewline ]

    pprAlt (tag, args, rhs)
      = iConcat [        iStr "<", iNum tag, iStr ">",
                         pprArgs args, iStr " -> ",
                         iIndent (pprExpr rhs) ]
pprExpr (ELam args body)
  = iConcat [  iStr "\\", pprArgs args, iStr ". ", iIndent (pprExpr body) ]
pprExpr (ELet isrec defns expr) 
= iConcat [ 	iStr keyword,
	iStr " ", iIndent (pprDefns defns),
	iStr "in ", (pprExpr expr)]
		where
	keyword | not isrec 	= "let"
		| isrec		= "letrec"
pprDefns :: [(Name,CoreExpr)] -> Iseq
pprDefns defns = iInterleave sep (map pprDefn defns)
                 where
                 sep = iConcat [ iStr ";", iNewline ]
pprDefn :: (Name, CoreExpr) -> Iseq
pprDefn (name, expr)
  = iConcat [ iStr name, iStr " = ", iIndent (pprExpr expr) ]
iConcat     :: [Iseq] -> Iseq
iInterleave :: Iseq -> [Iseq] -> Iseq
iConcat (x:xs) = x `iAppend` iConcat xs
iConcat []     = iNil
iInterleave i (x:[]) = x
iInterleave i (x:xs) = x `iAppend` (i `iAppend` iInterleave i xs)
iInterleave _ []     = iNil
pprProgram prog = iInterleave (iAppend (iStr " ;") iNewline) (map pprSc prog)
-- Isn't this what you meant?
pprSc (name, args, body)
  = iConcat [ iStr name, iSpace, pprArgs args,
              iStr " = ", iIndent (pprExpr body) ]
pprArgs args = iInterleave iSpace (map iStr args)
pprAExpr e | isAtomicExpr e = pprExpr e
pprAExpr e | otherwise = iConcat [iStr "(", pprExpr e, iStr ")"]
iSpace = iStr " "
iNil              = INil
iStr str 
|elem '\n' str = iAppend (iAppend (IStr (takeWhile (/= '\n') str)) INewline)
				 (iStr (tail (dropWhile (/= '\n') str)))
|otherwise 	   = IStr str
iAppend seq1 INil = seq1
iAppend INil seq2 = seq2
iAppend seq1 seq2 = IAppend seq1 seq2
data Iseq = INil
          | IStr String
          | IAppend Iseq Iseq
          | IIndent Iseq
          | INewline

iIndent seq = IIndent seq
iNewline    = INewline
instance Eq Iseq where
	INil		== INil		= True
	IStr a		== IStr b	= a == b
	IAppend a b	== IAppend c d	= a == c && b == d
	IIndent a	== IIndent b	= a == b
	INewline	== INewline	= True
_		== _		= False
flatten :: Int                       -- Current column; 0 for first column
            -> [(Iseq, Int)]         -- Work list
            -> String                -- Result
iDisplay seq = flatten 0 [(seq,0)]
flatten col ((INewline, indent) : seqs)
= '\n' : (take indent (repeat ' ')) ++ (flatten indent seqs)
flatten col ((IIndent seq, indent) : seqs)
	= flatten col ((seq, col) : seqs)
flatten _ [] = ""
flatten col ((IAppend seq1 seq2, indent) : seqs) 
= flatten col ((seq1, indent) : (seq2, indent) : seqs)
flatten col ((INil, indent) : seqs) = flatten col seqs
flatten col ((IStr s, indent) : seqs) = s ++ (flatten (col + (length s)) seqs)
iNum :: Int -> Iseq
iNum n = iStr (show n)
iFWNum :: Int -> Int -> Iseq
iFWNum width n
  = iStr (take (width - length digits) (repeat ' ') ++ digits)
    where
    digits = show n
iLayn :: [Iseq] -> Iseq
iLayn seqs = iConcat (map lay_item (zip [1..] seqs))
             where
             lay_item (n, seq)
               = iConcat [ iFWNum 4 n, iStr ") ", iIndent seq, iNewline ]
lexit :: String -> [String]
lexit s = cont (lex s)
where
cont [] = []
cont (([], []):_) = []
cont (([], b):_) = lexit b
cont ((a,[]):_) = [a]
cont ((a,b):_) = a : lexit b
syntax :: [Token] -> CoreProgram
parse :: String -> CoreProgram
parse = syntax . lexit
-- In Gofer I propose to compose this with some function
-- CoreProgram -> String, which will illustrate some sort of
-- execution machine, and then give this composition to catWith
-- from my utils
isIdChar, isWhiteSpace :: Char -> Bool
isIdChar c = isAlpha c || isDigit c || (c == '_')
isWhiteSpace c = c `elem` " \t\n"
twoCharOps :: [String]
twoCharOps = ["==", "~=", ">=", "<=", "->"]
type Token = String
lexCore :: String -> [Token]
lexCore ('-':'-':cs) = lexCore (dropWhile (/= '\n') cs)
lexCore (c1:'=':cs) 
| c1 `elem` "=~><" = [c1, '='] : lexCore cs
lexCore ('-':'>':cs) = "->":lexCore cs
lexCore (c:cs)
| isAlpha c = var_tok : lexCore rest1_cs
| isDigit c = num_token : lexCore rest2_cs
| isWhiteSpace c = lexCore cs
	| otherwise = [c] : lexCore cs
						where
						num_token = c : takeWhile isDigit cs
						rest2_cs = dropWhile isDigit cs
						var_tok = c : takeWhile isIdChar cs
						rest1_cs = dropWhile isIdChar cs
lexCore  [] = []
type Parser a = [Token] -> [(a, [Token])]
pLit :: String -> Parser String
pAlt :: Parser a -> Parser a -> Parser a
pAlt p1 p2 toks = (p1 toks) ++ (p2 toks)
pHelloOrGoodbye :: Parser String
pHelloOrGoodbye = (pLit "hello") `pAlt` (pLit "goodbye")
pThen :: (a -> b -> c) -> 
	Parser a -> Parser b -> Parser c
pThen combine p1 p2 toks
  = [ (combine v1 v2, toks2) | (v1,toks1) <- p1 toks,
                               (v2,toks2) <- p2 toks1]
pGreeting :: Parser (String, String)
pGreeting = pThen mk_pair pHelloOrGoodbye pVar
            where
            mk_pair hg name = (hg, name)
pThen3 :: (a -> b -> c -> d) -> 
Parser a -> Parser b -> Parser c -> Parser d
pThen3 comb p1 p2 p3 toks =
[(comb v1 v2 v3, toks')| (v1, toks1) <- p1 toks,
						 (v2, toks2) <- p2 toks1,
						 (v3, toks') <- p3 toks2]
pThen4 :: (a -> b -> c -> d -> e) -> 
Parser a -> Parser b -> Parser c -> Parser d -> Parser e
pThen4 comb p1 p2 p3 p4 toks =
	[(comb v1 v2 v3 v4, toks')| (v1, toks1) <- p1 toks,
							(v2, toks2) <- p2 toks1,
							(v3, toks3) <- p3 toks2,
							(v4, toks') <- p4 toks3]
pZeroOrMore :: Parser a -> Parser [a]
pGreetings :: Parser [(String, String)]
pGreetings = pZeroOrMore pGreeting
pZeroOrMore p = (pOneOrMore p) `pAlt` (pEmpty [])
failCont :: [([a], [Token])] -> [([a], [Token])] -> [([a], [Token])]
[] `failCont` q = q
p `failCont` q = p
pEmpty :: a -> Parser a
pOneOrMore :: Parser a -> Parser [a]
pEmpty x toks = [(x, toks)]
pOneOrMore p = pThen (:) p (pZeroOrMore p)
pGreetingsN :: Parser Int
pGreetingsN = (pZeroOrMore pGreeting) `pApply` length
pApply :: Parser a -> (a -> b) -> Parser b
pApply p f toks = [(f x, toks') | (x, toks') <- (p toks)]
pOneOrMoreWithSep :: Parser a -> Parser b -> Parser [a]
pOneOrMoreWithSep p psep = pThen (:) p (pOneOrMoreWithSep_c p psep)

pOneOrMoreWithSep_c p psep
  = (pThen discard_sep psep (pOneOrMoreWithSep p psep)) `pAlt`
    (pEmpty [])
    where
    discard_sep sep vs = vs
pSat :: (String -> Bool) -> Parser String
pLit s = pSat (== s)
pSat fp (s : toks) 
				| (fp s) == True = [(s, toks)]
				| otherwise 	 = []
pSat fp _ = []
keywords :: [String]
keywords = ["let", "letrec", "case", "in", "of", "Pack"]
pNum :: Parser Int
pVar = pSat isVar
       where
       isVar s = (isAlpha (head s) && s `notElem` keywords) || s == "_"
pNum = pSat (isDigit . head) `pApply` numval
numval :: String -> Int
numval = foldl (\a c -> 10 * a + ord c - ord '0') 0
numval8 :: String -> Int
numval8 = foldl (\a c -> 8 * a + ord c - ord '0') 0
syntax = take_first_parse . pProgram
where
take_first_parse ((prog,[]) : other_parses) = prog
take_first_parse ((p,x) : other) = 
pProgram :: Parser CoreProgram
pProgram = pThen keepFirst (pOneOrMoreWithSep pSc (pLit ";"))
						 (pZeroOrMore (pLit ";"))
where
keepFirst a _ = a
pSc :: Parser CoreScDefn
pSc = pThen4 mk_sc pVar (pZeroOrMore pVar) (pLit "=") pExpr
mk_sc :: Name -> [Name] -> Name -> CoreExpr -> (Name, [Name], CoreExpr)
mk_sc v1 vs _ expr = (v1, vs, expr)
pExpr = pLet `pAlt` (pCase `pAlt` (pLambda `pAlt` pExpr1))
pLet = pThen4 mk_let
              ((pLit "let") `pAlt` (pLit "letrec")) pDefns
              (pLit "in") pExpr
       where
       mk_let keyword defns in' expr = ELet (keyword == "letrec") defns expr

pDefns = pOneOrMoreWithSep pDefn (pLit ";")
pDefn = pThen3 mk_defn pVar (pLit "=") pExpr
        where
        mk_defn var equals rhs = (var,rhs)
pCase = pThen4 mk_case (pLit "case") pExpr (pLit "of") pAlters
        where
        mk_case case' e of' alts = ECase e alts

pAlters = pOneOrMoreWithSep pAlter (pLit ";")
pAlter = pThen4 mk_alt pTag (pZeroOrMore pVar) (pLit "->") pExpr
         where
         mk_alt tag args arrow rhs = (tag, args, rhs)

pTag = ((pChar `pApply` (mk_char . tail)) `pApply` get_charTag) `pAlt`
 (pThen3 get_tag (pLit "<") pNum (pLit ">"))
 where
 get_charTag (EConstr t _) = t
 get_tag lb tag rb = tag
pLambda = pThen4 mk_lam
              (pLit "\\") (pOneOrMore pVar) (pLit ".") pExpr
          where
          mk_lam lam vars dot expr = ELam vars expr
pExpr2 = pThen assembleOp pExpr3 pExpr2c
pExpr2c = (pThen FoundOp (pLit "&") pExpr2) `pAlt` (pEmpty NoOp)
pExpr3 = pThen assembleOp pExpr4 pExpr3c
pExpr3c = (pThen FoundOp pRelop pExpr4) `pAlt` (pEmpty NoOp)
pRelop = pSat (`elem` relops)
         where
         relops = ["<=", "<", ">=", ">", "==", "~="]
pExpr4 = pThen assembleOp pExpr5 pExpr4c
pExpr4c = (pThen FoundOp (pLit "+") pExpr4) `pAlt`
          ((pThen FoundOp (pLit "-") pExpr5) `pAlt`
           (pEmpty NoOp))
pExpr5 = pThen assembleOp pExpr6 pExpr5c
pExpr5c = (pThen FoundOp (pLit "*") pExpr5) `pAlt`
          ((pThen FoundOp (pLit "/") pExpr6) `pAlt`
           (pEmpty NoOp))
pExpr6 = (pOneOrMore pAtomic) `pApply` mk_ap_chain
         where
         mk_ap_chain (fn:args) = foldl EAp fn args
pAtomic = pConstr `pAlt`
	(pBracExpr `pAlt`
	(pUnion `pAlt`
	((pVar `pApply` EVar) `pAlt`
	((pNum `pApply` ENum) `pAlt`
	((pChar `pApply` (mk_char . tail)) `pAlt`
	((pString `pApply` (mk_string . tail))))))))
pBracExpr = pThen3 mk_brack (pLit "(") pExpr (pLit ")")
            where
            mk_brack open expr close = expr
pConstr = pThen4 pick_constr (pLit "Cons") (pLit "{") pTagArity (pLit "}")
          where
          pick_constr cons lbrack constr rbrack = constr
          pTagArity = pThen3 mk_constr pNum (pLit ",") pNum
          mk_constr tag comma arity = EConstr tag arity
pUnion = pThen3 mk_union (pLit "<") (pOneOrMoreWithSep pExpr (pLit ",")) (pLit ">")
	where mk_union langle exprs rangle = EUnion exprs
pChar = pSat isChar
    where
    isChar s = (head s) == '\''

mk_char ('\\':'^':c:as) = EConstr ((ord c) - 64) 0
mk_char ('\\':'o':as) = EConstr (numval8 as) 0
mk_char ('\\':'n':as) = EConstr 10 0
mk_char ('\\':'r':as) = EConstr 13 0
mk_char ('\\':'\\':as) = EConstr (ord '\\') 0
mk_char ('\\':'\'':as) = EConstr (ord '\'') 0
mk_char (a:as) = EConstr (ord a) 0

pString = pSat isString
    where
    isString "" = False
    isString s = (head s) == '\"'

mk_string "\"" = EConstr 1 0
mk_string s1@(s:ss) = EAp (EAp (EConstr 2 2) (mk_char s1)) (mk_string ss)
data PartialExpr = NoOp | FoundOp Name CoreExpr
pExpr1c :: Parser PartialExpr
pExpr1c = (pThen FoundOp (pLit "|") pExpr1) `pAlt` (pEmpty NoOp)
pExpr1 :: Parser CoreExpr
pExpr1 = pThen assembleOp pExpr2 pExpr1c
assembleOp :: CoreExpr -> PartialExpr -> CoreExpr
assembleOp e1 NoOp = e1
assembleOp e1 (FoundOp op e2) = EAp (EAp (EVar op) e1) e2
