>

\input Main

> module Impl where
> import Utils
> import AbsSyntax
> import Show(showResults, showShortResults, showAnalyserResults, 
>       showHiLevel)
> import Compiler(compile)
> import Evaluator(eval, neq, resourcesExhausted)
> import StateType

\section{The $G^\#$-maschine}
\label{minimal-gm}

\subsection{Basic structure\index{G-machine@@$G^\#$-machine}}

On first sight the $G^\#$-machine 
\index{G-machine@@$G^\#$-machine} presents itself like its concrete
counterpart. There are three distinct top level calls each with a
different degree of verbosity. 

We have:

\begin{itemize} 
\item The function @hrun@ (stands for \underline{h}igh-level
\underline{run}) which shows the expression after each reduction
step.\index{expression!showing the}

> hrun :: [Char] -> [Char]
> hrun = showHiLevel . eval . compile . parse

\item The function @drun@ (stands for \underline{d}etailed
\underline{run}) shows in detail every state of the abstract machine.

> drun :: [Char] -> [Char]
> drun = showResults . eval . compile . parse

\item The function @crun@ (stands for \underline{c}ompile and
\underline{run}) only shows the final state.

> crun :: [Char] -> [Char]
> crun = showShortResults . eval . compile . parse

\end{itemize}

\subsection{Fundamental datatypes}

\input StateType

\subsection{The evaluator\index{G-machine@@$G^\#$-machine}}
\label{gm:ss:eval1}
\input Evaluator

\section{Strictness analysis\index{strictness analysis}} 

To analyse strictness in the $i$th argument of a function, we start a
reduction\index{reduction!abstract} with all the arguments set to
$\top$\index{$\top$} except for the $i$th argument which is set to
$\bot$.\index{$\bot$} This reduction\index{reduction!abstract} might
exhaust our resources. If the reduction does not return 
$\bot$\index{$\bot$} within our resource constraints, we assume the
function not to be strict in its $i$th argument. This assumption is safe.

The following function takes a state, the name of a function and an
argument position to obtain a list of starting states. These starting
states are then given to @eval@ for evaluation. If @eval@ returns a
state which meets our resource constraints and which has
$\bot$\index{$\bot$} on top of its stack, the function is strict in
its $i$th argumant.

> analyserStartStates :: GmState -> Name -> Int -> [GmState]
> analyserStartStates state f i 
>   = [putCode (ins "Bot") state] -- , putCode (ins "Top") state]
>   where   
>   (NGlobal n _) = rLookup (getHeap state) 
>                       (aLookup (getGlobals state) f 
>                       (error ("Undefined Global " ++ f)))
>   ins x = take (n-i) (repeat (Pushglobal "Top")) ++
>           [Pushglobal x] ++
>           take (i-1) (repeat (Pushglobal "Top")) ++
>           [Pushglobal f] ++
>           take n (repeat Mkap) ++
>           [Eval]

When a function has been analysed we want to store the information
gained, to use it in analysing other functions. For this purpose we
use the a trick which was also inspired by the fundamental idea of the
G-machine. We compile the strictness information\index{strictness
information} to $G^\#$-code\index{$G^\#$-code} which, when executed
has the effect of applying the strictness information.
To apply the strictness information within the analyser means to
return $\bot$\index{$\bot$} for a function, if $\bot$\index{$\bot$}
can be found in a known-to-be-strict position.
So for the strict positions the following has to be done: the argument
has to be evaluated and if it evaluates to $\bot$\index{$\bot$} we
immediately return from the evaluation of the function with
$\bot$\index{$\bot$} on top of the stack. If we do not find
$\bot$\index{$\bot$} in any strict position we call the original function.
Because the roots of the redexes we evaluate are overwritten the
argumants do not have to be evaluated again in the body of the
original function. (More precisely: to evaluate them we only have to
llok up the value the root was indirected to.) So we do not have much
of a performance loss as compared to evaluation without considering
strictness information.

> analyserCode :: Int -> [Bool] -> Name -> GmCode
> analyserCode arg [] f =   [Pushglobal ('_':f)] ++
>               take arg (repeat Mkap) ++
>               [Update 0, Unwind]
>
> analyserCode arg (False:xs) f = analyserCode (arg+1) xs f
> analyserCode arg (True:xs) f  = 
>   [Push arg,
>    Eval,
>    Pushglobal "Bot",
>    AbsEq,
>    Casejump [(2:= [Pushglobal "Bot",
>                    Update (arg+2+length xs),
>                    Pop (arg+2+length xs),
>                    Unwind]),
>              (1:= ((Pop 1):(analyserCode (arg+1) xs f)))]]

In @strictPosition@ we have the function which takes the list of
starting states passes these on to @eval@ and tests if the result is
$\bot$\index{$\bot$} and if it was found within the resource
constraints.

> strictPosition :: [GmState] -> [Bool]
> strictPosition [] = []
> strictPosition (x:xs) = 
>   let e = last (eval x); hp = getHeap e in 
>       [not (resourcesExhausted e) && 
>       (neq hp (rLookup hp (head (getStack e))) NBot)]
>       ++ strictPosition xs

@strictPositions@ applies @strictPosition@ to all the argument
positions of a function.

> strictPositions :: GmState -> Name -> [Bool]
> strictPositions state f = 
>   map (\z -> any (True ==) 
>       (strictPosition (analyserStartStates state f z))) [1..n]
>   where   (NGlobal n _) = rLookup (getHeap state) 
>                  (aLookup (getGlobals state) f 
>                   (error ("Undefined Global " ++ f)))

\par

@analyseFunction@ analyses {\em one} function given its name.

> analyseFunction :: Name -> GmState -> GmState
> analyseFunction f state = 
>   putGlobals ((('_':f) := olda):(aUpdate gl f newa))
>       (putHeap newh 
>       (putOutput 
>           (iConcat [op, delim, 
>                 iStr f, iStr " [", 
>                 iInterleave (iStr ", ") 
>                   (map (iStr . myshow)  striP), 
>                 iStr "]"]) 
>           state))
>   where   (newh, newa) = bAlloc hp (NGlobal n 
>                               (analyserCode 0 striP f))
>           olda = aLookup gl f 
>                       (error ("Undefined Global " ++ f))
>           (NGlobal n _) = rLookup hp olda
>           hp = getHeap state
>           gl = getGlobals state
>           striP = strictPositions state f
>           op = getOutput state
>           delim 
>               | op /= iNil = iNewline
>               | otherwise  = iNil
>           myshow True = "strict"
>           myshow False = "?"

@analyseFunctions@ analyses {\em all} functions given in a list.

> analyseFunctions :: [Name] -> GmState -> GmState
> analyseFunctions [] state = state
> analyseFunctions (x:xs) state 
>   = analyseFunctions xs st
>   where 
>       st = analyseFunction x state 


To obtain a list of all function names in a file we use the function
@allNames@. 

> allNames :: GmState -> [Name]
> allNames st = [nm | (nm := a) <- getGlobals st, 
>           (NGlobal n _) <- [rLookup (getHeap st) a],
>            n > 0]

\section{Dependency analysis\index{Dependency analysis}}

To analyse the dependency of functions we need a means to find all the
functions called from a $G^\#$-code fragment.

> calledNamesFromCode :: GmCode -> [Name]
> calledNamesFromCode (Pushglobal f : xs) 
>   = f : calledNamesFromCode xs
> calledNamesFromCode (Casejump cs : xs)  
>   = (concat [calledNamesFromCode x | _ := x <- cs])
>       ++ calledNamesFromCode xs
> calledNamesFromCode (_:xs) = calledNamesFromCode xs
> calledNamesFromCode []     = []

\par

> calledNames :: Node -> [Name]
> calledNames (NGlobal n cs)    = calledNamesFromCode cs
> calledNames _         = []

We need types for directed graphs and functions to evaluate when
visiting a node. Let's turn towards graphs first. 

> data GraphEdge a = a :=> a
> type CallGraphEdge = GraphEdge Name
> type Graph a b = (AssocList a (Bool, b), [GraphEdge a])
> type CallGraph = Graph Name Int
> type SCC a = [a]
> type CallSCC = SCC Name

We define an order on the graphs based on the numbers of the nodes.


Now we can obtain the graph of the calling structure with @callGraph@.

> callGraph :: GmState -> CallGraph
> callGraph s = (v, e)
>   where
>       v = [n := (False,0)|(n := _) <- glob]
>       e = [x :=> y| (x := a) <- glob, 
>               y <- nub (calledNames (rLookup (getHeap s) a))]
>       glob = getGlobals s

To reverse a graph the source and destination of the arcs are swapped.

> reverseGraph :: Graph a b -> Graph a b
> reverseGraph (v,e) = (v,[y:=>x|x:=>y <- e])

The strongly connected components can then be found with dfs from the
largets remaining node.

\begin{verbatim}
 findSCCs :: Ord b => Graph a b -> [SCC a] -> [SCC a]
 findSCCs ([],_) s = s
 findSCCs (v,e)  s = findSCCs g (sc:s)
    where 
        (n := _) = maximum v
        (sc, g) = dfsVisitEarly n (v,e) ([],(v,e))
\end{verbatim}

We distinguish between @dfsVisitEarly@ and @dfsVisitLate@. In the
former we collect the nodes in the sequence they are found and in the
latter we collect a node after collecting its children. 

> dfsVisitEarly :: Eq a => a -> Graph a b -> ([a], Graph a b) 
>           -> ([a], Graph a b)
> dfsVisitEarly n (v,e) (sc,r) = 
>   dfss dfsVisitEarly [y | (x :=> y) <- e', x == n] (v',e') 
>       (n:sc, (v',e'))
>   where
>       v' = [n' := (x,y)|(n' := (x,y)) <- v, n' /= n]
>       e' = [x :=> y|(x :=> y) <- e, y /= n]

\par

> dfsVisitLate :: Eq a => a -> Graph a b -> ([a], Graph a b) 
>           -> ([a], Graph a b)
> dfsVisitLate n (v,e) (s,r) = (n:sc, res)
>   where
>   (sc, res) = dfss dfsVisitLate [y | (x :=> y) <- e', x == n] 
>       (v',e') (s,(v',e'))
>   v' = [n' := (x,y)|(n' := (x,y)) <- v, n' /= n]
>   e' = [x :=> y|(x :=> y) <- e, y /= n]

\par

> dfss :: ( a -> Graph a b -> ([a], Graph a b) 
>           -> ([a], Graph a b)) -> [a] -> Graph a b
>           -> ([a], Graph a b) -> ([a], Graph a b)
> dfss dfs [] _ r = r
> dfss dfs (n:ns) g r = dfss dfs ns r' (sc',r')
>   where
>       (sc',r') = dfs n g r

\par

> sccNumbering :: [Name] -> Int -> [(Name, Int)]
> sccNumbering [] _ = []
> sccNumbering (n:ns) i = (n,i) : (sccNumbering ns (i-1))

\par

\begin{verbatim}
 callGraphSCCs :: CallGraph -> [CallSCC]
 callGraphSCCs callGraph@(v,e) = 
    findSCCs (reverseGraph ([n := (b,i)|(n := (b,i)) <- v', 
                i /= 0, n /= "Top", n /= "Bot", 
                aLookup v' ('_':n) (False,0) == (False,0)],e)) []
    where
    (nameList,_) = dfsVisitLate "main" callGraph ([], callGraph)
    v' =  foldl (\x (y,z) -> aUpdate x y (False, z)) v 
            (sccNumbering nameList (length nameList))
\end{verbatim}
\par

\section{The compiler\index{G-machine!compiler}}
\input CompilerTypes
\input Compiler

\section{Output of results}

\input Show

%end of this file
